# Generated by Django 4.2.11 on 2024-07-08 09:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0051_merge_20240702_1220"),
    ]

    operations = [
        migrations.AlterField(
            model_name="searchmodelconfig",
            name="bi_encoder_docs_encode_config",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="searchmodelconfig",
            name="bi_encoder_model_config",
            field=models.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name="searchmodelconfig",
            name="bi_encoder_query_encode_config",
            field=models.JSONField(blank=True, default=dict),
        ),
    ]
