# Generated by Django 5.0.10 on 2025-01-20 16:33

from typing import List

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0079_entry_file_object"),
        ("database", "0080_speechtotextmodeloptions_ai_model_api"),
    ]

    operations: List[str] = []
